#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "FW-MAP.h"

// Example of using FWMAP:
//
// Minimize function   f(x) = \sum_{i=0}^{node_num-2} f_i(x)
// where x has 'node_num' variables with x(i) \in {0,1,...,K-1} and term f_i depends on variables x(i) and x(i+1)
//
// The problem must be encoded using {0,1}-valued indicator variables. In total, there will be 'node_num*K' such variables
//
// Three ways of creating 'FWMAP' are demonstrated:
// Create_NoMapping():      assume that term f_i depends on all 'node_num' variables. (Very inefficient relaxation!)
// Create_Mapping():        explicitly encode that term f_i depends only on variables x(i) and x(i+1)
// Create_MappingCompact(): same as above by additionally use compact representation of labelings

struct ChainProblem
{
	typedef int REAL;

	ChainProblem(int node_num, int K); // K = # labels
	~ChainProblem();
	REAL Minimize(); // minimize exactly by dynamic programming

	struct Edge
	{
		ChainProblem* P;
		REAL* costs;
		int i, j;
	};

	///////////////////////////////////////////////////////////////////////////////

	FWMAP* Create_NoMapping();
	static double _min_fn_NoMapping(double* lambdai, YPtr _y, TermData term_data)
	{
		Edge* e = (Edge*) term_data;
		return e->P->min_fn_NoMapping(lambdai, (double*)_y, e);
	}
	double min_fn_NoMapping(double* lambdai, double* y, Edge* e);

	///////////////////////////////////////////////////////////////////////////////

	FWMAP* Create_Mapping();
	static double _min_fn_Mapping(double* lambdai, YPtr _y, TermData term_data)
	{
		Edge* e = (Edge*) term_data;
		return e->P->min_fn_Mapping(lambdai, (double*)_y, e);
	}
	double min_fn_Mapping(double* lambdai, double* y, Edge* e);

	///////////////////////////////////////////////////////////////////////////////

	FWMAP* Create_MappingCompact();
	static double _min_fn_MappingCompact(double* lambdai, YPtr _y, TermData term_data)
	{
		Edge* e = (Edge*) term_data;
		return e->P->min_fn_MappingCompact(lambdai, (int*)_y, e);
	}
	static void _copy_fn_MappingCompact(double* xi, YPtr _y, TermData term_data)
	{
		Edge* e = (Edge*) term_data;
		e->P->copy_fn_MappingCompact(xi, (int*)_y);
	}
	static double _dot_product_fn_MappingCompact(double* lambdai, YPtr _y, TermData term_data)
	{
		Edge* e = (Edge*) term_data;
		return e->P->dot_product_fn_MappingCompact(lambdai, (int*)_y);
	}
	double min_fn_MappingCompact(double* lambdai, int* y, Edge* e);
	void copy_fn_MappingCompact(double* xi, int* y);
	double dot_product_fn_MappingCompact(double* lambdai, int* y);

	////////////////////////////////////////////////////

	int node_num, K, edge_num;
	Edge* edges;
	Buffer buf;
};


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

ChainProblem::ChainProblem(int _node_num, int _K) : node_num(_node_num), edge_num(_node_num-1), K(_K), buf(1024)
{
	int i, j;
	edges = new Edge[node_num-1];
	for (i=0; i<node_num-1; i++)
	{
		Edge* e = edges + i;
		e->P = this;
		e->i = i;
		e->j = i+1;
		e->costs = new REAL[K*K];
		for (j=0; j<K*K; j++)
		{
			e->costs[j] = rand() % 5;
		}
	}
}

ChainProblem::~ChainProblem()
{
	int i;
	for (i=0; i<node_num-1; i++)
	{
		Edge* e = edges + i;
		delete [] e->costs;
	}
	delete [] edges;
}

ChainProblem::REAL ChainProblem::Minimize()
{
	int i, a, b;
	REAL* A = new REAL[2*K];
	REAL* B = A + K;

	memset(A, 0, K*sizeof(REAL));
	for (i=0; i<edge_num; i++)
	{
		REAL* costs = edges[i].costs;
		for (b=0; b<K; b++)
		{
			B[b] = A[0] + costs[b];
			for (a=1; a<K; a++)
			{
				if (B[b] > A[a] + costs[a*K + b]) B[b] = A[a] + costs[a*K + b];
			}
		}
		memcpy(A, B, K*sizeof(REAL));
	}
	REAL v = A[0];
	for (a=1; a<K; a++)
	{
		if (v > A[a]) v = A[a];
	}

	delete [] A;
	return v;
}



////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

FWMAP* ChainProblem::Create_NoMapping()
{
	FWMAP* F = new FWMAP(node_num*K, edge_num, _min_fn_NoMapping, NULL, NULL);
	int e;
	for (e=0; e<edge_num; e++)
	{
		F->SetTerm(e, &edges[e], node_num*K);
	}
	return F;
}

double ChainProblem::min_fn_NoMapping(double* lambdai, double* y, Edge* e)
{
	memset(y, 0, node_num*K*sizeof(double));

	int i, xi, xj, xi_best=-1, xj_best=-1;
	double cost_best = 0;
	for (xi=0; xi<K; xi++)
	for (xj=0; xj<K; xj++)
	{
		double cost = e->costs[xi*K + xj] + lambdai[e->i*K + xi] + lambdai[e->j*K + xj];
		if (xi_best < 0 || cost_best > cost)
		{
			cost_best = cost;
			xi_best = xi;
			xj_best = xj;
		}
	}
	y[e->i*K + xi_best] = 1;
	y[e->j*K + xj_best] = 1;
	double sum = e->costs[xi_best*K + xj_best];

	for (i=0; i<node_num; i++)
	{
		if (i==e->i || i==e->j) continue;
		for (xi=0; xi<K; xi++)
		{
			if (xi==0 || cost_best > lambdai[i*K + xi])
			{
				cost_best = lambdai[i*K + xi];
				xi_best = xi;
			}
		}
		y[i*K + xi_best] = 1;
	}

	return sum;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

FWMAP* ChainProblem::Create_Mapping()
{
	FWMAP* F = new FWMAP(node_num*K, edge_num, _min_fn_Mapping, NULL, NULL);
	int e, a;
	for (e=0; e<edge_num; e++)
	{
		int* mapping = (int*) buf.Alloc(2*K*sizeof(int));
		for (a=0; a<K; a++)
		{
			mapping[a]   = edges[e].i*K + a;
			mapping[K+a] = edges[e].j*K + a;
		}
		F->SetTerm(e, &edges[e], 2*K, mapping);
	}
	return F;
}

double ChainProblem::min_fn_Mapping(double* lambdai, double* y, Edge* e)
{
	memset(y, 0, 2*K*sizeof(double));

	int xi, xj, xi_best=-1, xj_best=-1;
	double cost_best = 0;
	for (xi=0; xi<K; xi++)
	for (xj=0; xj<K; xj++)
	{
		double cost = e->costs[xi*K + xj] + lambdai[xi] + lambdai[K + xj];
		if (xi_best < 0 || cost_best > cost)
		{
			cost_best = cost;
			xi_best = xi;
			xj_best = xj;
		}
	}
	y[    xi_best] = 1;
	y[K + xj_best] = 1;
	return e->costs[xi_best*K + xj_best];
}

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

FWMAP* ChainProblem::Create_MappingCompact()
{
	FWMAP* F = new FWMAP(node_num*K, edge_num, _min_fn_MappingCompact, _copy_fn_MappingCompact, _dot_product_fn_MappingCompact);
	int e, a;
	for (e=0; e<edge_num; e++)
	{
		int* mapping = (int*) buf.Alloc(2*K*sizeof(int));
		for (a=0; a<K; a++)
		{
			mapping[a]   = edges[e].i*K + a;
			mapping[K+a] = edges[e].j*K + a;
		}
		F->SetTerm(e, &edges[e], 2*K, mapping, 2*sizeof(int));
	}
	return F;
}

double ChainProblem::min_fn_MappingCompact(double* lambdai, int* y, Edge* e)
{
	int xi, xj, xi_best=-1, xj_best=-1;
	double cost_best = 0;
	for (xi=0; xi<K; xi++)
	for (xj=0; xj<K; xj++)
	{
		double cost = e->costs[xi*K + xj] + lambdai[xi] + lambdai[K + xj];
		if (xi_best < 0 || cost_best > cost)
		{
			cost_best = cost;
			xi_best = xi;
			xj_best = xj;
		}
	}
	y[0] = xi_best;
	y[1] = xj_best;
	return e->costs[xi_best*K + xj_best];
}

void ChainProblem::copy_fn_MappingCompact(double* xi, int* y)
{
	memset(xi, 0, 2*K*sizeof(double));
	xi[    y[0]] = 1;
	xi[K + y[1]] = 1;
}

double ChainProblem::dot_product_fn_MappingCompact(double* lambdai, int* y)
{
	return lambdai[y[0]] + lambdai[K + y[1]];
}

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////


int main()
{
	unsigned seed;
	for (seed=1; ; seed++)
	{
		srand(seed);
		printf("seed=%d\n", seed);

		ChainProblem* P = new ChainProblem(10, 3);

		///////////////////////

		int flag;
		for (flag=0; flag<3; flag++)
		{
			FWMAP* F;
			if      (flag == 0) F = P->Create_NoMapping();
			else if (flag == 1) F = P->Create_Mapping();
			else if (flag == 2) F = P->Create_MappingCompact();

			F->options.gap0_max = 1e-3; // can set some parameter here
			F->Solve();
			printf("MAP=%f\n", (double)P->Minimize());

			delete F;
			getchar();
		}

		///////////////////////

		delete P;
	}

	return 0;
}

